#include "darwin_robot.h"
#include "darwin_robot_exceptions.h"
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef _HAVE_XSD
#include "xml/darwin_config.hxx"
#endif

const std::string servo_names[MAX_NUM_SERVOS]={std::string("Servo0"),
                                               std::string("j_shoulder_r"),
                                               std::string("j_shoulder_l"),
                                               std::string("j_high_arm_r"),
                                               std::string("j_high_arm_l"),
                                               std::string("j_low_arm_r"),
                                               std::string("j_low_arm_l"),
                                               std::string("j_pelvis_r"),
                                               std::string("j_pelvis_l"),
                                               std::string("j_thigh1_r"),
                                               std::string("j_thigh1_l"),
                                               std::string("j_thigh2_r"),
                                               std::string("j_thigh2_l"),
                                               std::string("j_tibia_r"),
                                               std::string("j_tibia_l"),
                                               std::string("j_ankle1_r"),
                                               std::string("j_ankle1_l"),
                                               std::string("j_ankle2_r"),
                                               std::string("j_ankle2_l"),
                                               std::string("j_pan"),
                                               std::string("j_tilt"),
                                               std::string("Servo21"),
                                               std::string("Servo22"),
                                               std::string("Servo23"),
                                               std::string("Servo24"),
                                               std::string("Servo25"),
                                               std::string("Servo26")};

const led_info_t leds[GPIO_NUM_LEDS]={{DARWIN_TX_LED_CNTRL,DARWIN_TX_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_RX_LED_CNTRL,DARWIN_RX_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_PLAY_LED_CNTRL,DARWIN_PLAY_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_EDIT_LED_CNTRL,DARWIN_EDIT_LED_PERIOD_L,GPIO_VALUE,GPIO_TOGGLE,GPIO_BLINK},
                                      {DARWIN_MNG_LED_CNTRL,0x0000,GPIO_VALUE,GPIO_TOGGLE,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_R,GPIO_TOGGLE_R,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_G,GPIO_TOGGLE_G,0x00},
                                      {DARWIN_AUX1_LED_CNTRL,DARWIN_AUX1_LED_COLOR_L,GPIO_VALUE_B,GPIO_TOGGLE_B,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_R,GPIO_TOGGLE_R,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_G,GPIO_TOGGLE_G,0x00},
                                      {DARWIN_AUX2_LED_CNTRL,DARWIN_AUX2_LED_COLOR_L,GPIO_VALUE_B,GPIO_TOGGLE_B,0x00}};


CDarwinRobot::CDarwinRobot(const std::string &name,std::string &bus_id,int bus_speed, unsigned char id)
{
  int num_buses;

  this->robot_device=NULL;
  if(name!="")
  {
    this->robot_name=name;
    this->dyn_server=CDynamixelServerFTDI::instance();
    num_buses=this->dyn_server->get_num_buses();
    if(num_buses>0)
    {
      this->dyn_server->config_bus(bus_id,bus_speed);
      this->robot_device=dyn_server->get_device(id,dyn_version2);
      this->robot_id=id;
    }
    else
    {
      /* handle exceptions */
      throw CDarwinRobotException(_HERE_,"No FTDI USB devices found");
    }
  }
  else
  {
    /* handle exceptions */    
    throw CDarwinRobotException(_HERE_,"Invalid robot name");
  }
}

// GPIO interface
void CDarwinRobot::gpio_set_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,leds[led].value_mask);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_clear_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,0x00);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_toggle_led(led_t led)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(leds[led].control_reg,leds[led].toggle_mask);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_blink_led(led_t led,int period_ms)
{
  unsigned char data[3];

  if(this->robot_device!=NULL)
  {
    if(led>=LED_TX && led<=LED_3)
    {
      data[0]=leds[led].blink_mask;// blink led
      data[1]=period_ms%256;
      data[2]=period_ms/256;
      this->robot_device->write_registers(leds[led].control_reg,data,3);
    }
    else
      throw CDarwinRobotException(_HERE_,"The desired LED can not be configured to blink");
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_aux1_color(double red, double green, double blue)
{
  unsigned char data[3];
  unsigned short int color;

  if(this->robot_device!=NULL)
  {
    data[0]=GPIO_COLOR;
    /* saturate the color */
    if(red>1.0) red=1.0;
    else if(red<0.0) red=0.0;
    if(green>1.0) green=1.0;
    else if(green<0.0) green=0.0;
    if(blue>1.0) blue=1.0;
    else if(blue<0.0) blue=0.0;
    color=((unsigned short int)(red*31.0))+(((unsigned short int)(green*31.0))<<5)+(((unsigned short int)(blue*31.0))<<10);
    data[1]=color%256;
    data[2]=color/256;
    this->robot_device->write_registers(DARWIN_AUX1_LED_CNTRL,data,3);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::gpio_aux2_color(double red, double green, double blue)
{
  unsigned char data[3];
  unsigned short int color;

  if(this->robot_device!=NULL)
  {
    data[0]=GPIO_COLOR;
    /* saturate the color */
    if(red>1.0) red=1.0;
    else if(red<0.0) red=0.0;
    if(green>1.0) green=1.0;
    else if(green<0.0) green=0.0;
    if(blue>1.0) blue=1.0;
    else if(blue<0.0) blue=0.0;
    color=((unsigned short int)(red*31.0))+(((unsigned short int)(green*31.0))<<5)+(((unsigned short int)(blue*31.0))<<10);
    data[1]=color%256;
    data[2]=color/256;
    this->robot_device->write_registers(DARWIN_AUX2_LED_CNTRL,data,3);
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::is_button_pressed(pushbutton_t pushbutton)
{
  unsigned char value;

  if(this->robot_device!=NULL)
  {
    switch(pushbutton)
    {
      case START_PB: this->robot_device->read_byte_register(DARWIN_START_PB_CNTRL,&value);
                     break;
      case MODE_PB: this->robot_device->read_byte_register(DARWIN_MODE_PB_CNTRL,&value);
                    break;
    }
    if(value&GPIO_VALUE)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// ADC interface
void CDarwinRobot::adc_start(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_CNTRL,ADC_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::adc_is_running(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_ADC_CNTRL,&status);
    if(status&ADC_RUNNING)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::adc_set_period(unsigned char period_ms)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_PERIOD,period_ms);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}
double CDarwinRobot::adc_get_value(adc_t adc)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_CH1_L+((int)adc)*2,&value);

    return ((double)value)/((double)(1<<12));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::adc_get_temperature(void)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_TEMP_L,&value);
    std::cout << std::hex << value << std::endl;

    return ((double)value)/((double)(1<<10));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

double CDarwinRobot::adc_get_vrefint(void)
{
  unsigned short int value;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_ADC_VREF_L,&value);

    return ((double)value)/((double)(1<<12));
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::adc_stop(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_ADC_CNTRL,ADC_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// imu interface
void CDarwinRobot::imu_start(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_START);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_set_cal_samples(unsigned short int num_samples)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_word_register(DARWIN_IMU_CAL_SAMPLES_L,num_samples);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

unsigned short int CDarwinRobot::imu_get_cal_samples(void)
{
  unsigned short int samples;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_word_register(DARWIN_IMU_CAL_SAMPLES_L,&samples);
    return samples;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_stop(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_STOP);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_start_gyro_cal(void)
{
  if(this->robot_device!=NULL)
    this->robot_device->write_byte_register(DARWIN_IMU_CNTRL,IMU_START_CAL);
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_gyro_cal_done(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_CALIBRATING)
      return true;
    else
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_accel_detected(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_ACCEL_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

bool CDarwinRobot::imu_is_gyro_detected(void)
{
  unsigned char status;

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_byte_register(DARWIN_IMU_CNTRL,&status);
    if(status&IMU_GYRO_DET)
      return true;
    else 
      return false;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_get_accel_data(double *x,double *y,double *z)
{
  unsigned char data[6];

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_IMU_ACCEL_X_L,data,6);

    *x=((double)((signed short int)(data[0]+(data[1]<<8))))/1000.0;
    *y=((double)((signed short int)(data[2]+(data[3]<<8))))/1000.0;
    *z=((double)((signed short int)(data[4]+(data[5]<<8))))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

void CDarwinRobot::imu_get_gyro_data(double *x,double *y,double *z)
{
  unsigned char data[6];

  if(this->robot_device!=NULL)
  {
    this->robot_device->read_registers(DARWIN_IMU_GYRO_X_L,data,6);

    *x=((double)(data[0]+(data[1]<<8)))/1000.0;
    *y=((double)(data[2]+(data[3]<<8)))/1000.0;
    *z=((double)(data[4]+(data[5]<<8)))/1000.0;
  }
  else
    throw CDarwinRobotException(_HERE_,"Invalid robot device");
}

// motion manager interface
/*void CDarwinRobot::mm_set_period(double period_ms)
{
  unsigned short period;
  // internal time in 0|16 fixed point float format in seconds
  period=(period_ms/1000.0)*65536;
  this->robot_device->write_word_register(DARWIN_MM_PERIOD_LOW,period);
}

double CDarwinRobot::mm_get_period(void)
{
  unsigned short period;

  this->robot_device->read_word_register(DARWIN_MM_PERIOD_LOW,&period);
  return (((double)period)*1000.0)/65536;
}

double CDarwinRobot::mm_get_current_period(void)
{
  unsigned short period;

  this->robot_device->read_word_register(DARWIN_MM_CUR_PERIOD,&period);
  return ((double)period)/1000.0;
}

void CDarwinRobot::mm_get_num_servos(int *v1_servos,int *v2_servos)
{
  unsigned char data;

  this->robot_device->read_byte_register(DARWIN_MM_NUM_SERVOS,&data);
  *v1_servos=data&0x1F;
  *v2_servos=(data>>5)&0x07;
}

void CDarwinRobot::mm_start(void)
{
  unsigned char status;

  this->robot_device->read_byte_register(DARWIN_MM_CONTROL,&status);
  status|=0x01;
  this->robot_device->write_byte_register(DARWIN_MM_CONTROL,status);
}

void CDarwinRobot::mm_stop(void)
{
  unsigned char status;

  this->robot_device->read_byte_register(DARWIN_MM_CONTROL,&status);
  status&=0xFE;
  this->robot_device->write_byte_register(DARWIN_MM_CONTROL,status);
}

void CDarwinRobot::mm_enable_power(void)
{
  this->robot_device->write_byte_register(DARWIN_DXL_POWER,0x01);
}

void CDarwinRobot::mm_disable_power(void)
{
  this->robot_device->write_byte_register(DARWIN_DXL_POWER,0x00);
  sleep(1);
}

bool CDarwinRobot::mm_is_power_enabled(void)
{
  unsigned char value;

  this->robot_device->read_byte_register(DARWIN_DXL_POWER,&value);
  if(value==0x01)
    return true;
  else
    return false;
}

void CDarwinRobot::mm_enable_servo(unsigned char servo_id)
{
  unsigned char value;

  if(servo_id>MAX_NUM_SERVOS)
  {
    // handle exceptions
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  // get the current value
  this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
  if(servo_id%2)// odd servo
    value|=0x08;
  else// even servo
    value|=0x80;
  this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
}

void CDarwinRobot::mm_disable_servo(unsigned char servo_id)
{
  unsigned char value;

  if(servo_id>MAX_NUM_SERVOS)
  {
    // handle exceptions 
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  // get the current value
  this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
  if(servo_id%2)// odd servo
    value&=0xF7;
  else// even servo
    value&=0x7F;
  this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
}

bool CDarwinRobot::mm_is_servo_enabled(unsigned char servo_id)
{
  unsigned char value;

  if(servo_id>MAX_NUM_SERVOS)
  {
    // handle exceptions
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  // get the current value
  this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
  if(servo_id%2)// odd servo
  {
    if(value&0x01)
      return true;
    else
      return false;
  }
  else// even servo
  {
    if(value&0x10)
      return true;
    else
      return false;
  }
}

void CDarwinRobot::mm_enable_balance(void)
{
  unsigned char status;

  this->robot_device->read_byte_register(DARWIN_MM_CONTROL,&status);
  status|=0x02;
  this->robot_device->write_byte_register(DARWIN_MM_CONTROL,status);
}

void CDarwinRobot::mm_disable_balance(void)
{
  unsigned char status;

  this->robot_device->read_byte_register(DARWIN_MM_CONTROL,&status);
  status&=0xFD;
  this->robot_device->write_byte_register(DARWIN_MM_CONTROL,status);
}

bool CDarwinRobot::mm_is_balance_enabled(void)
{
  unsigned char value;

  this->robot_device->read_byte_register(DARWIN_MM_STATUS,&value);
  if(value==0x01)
    return true;
  else
    return false;
}


void CDarwinRobot::mm_assign_module(unsigned char servo_id, mm_mode_t mode)
{
  unsigned char value;

  if(servo_id>MAX_NUM_SERVOS)
  {
    // handle exceptions
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  // get the current value
  this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
  if(servo_id%2)// odd servo
  {
    value&=0xF8;
    value|=(unsigned char)mode;
  }
  else// even servo
  {
    value&=0x8F;
    value|=(((unsigned char)mode)<<4);
  }
  this->robot_device->write_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,value);
}

void CDarwinRobot::mm_assign_module(std::string &servo,std::string &module)
{
  unsigned int i=0;

  for(i=0;i<MAX_NUM_SERVOS;i++)
  {
    if(servo==servo_names[i])
    {
      if(module=="action")
        this->mm_assign_module(i,DARWIN_MM_ACTION);
      else if(module=="walking")
        this->mm_assign_module(i,DARWIN_MM_WALKING);
      else if(module=="joints")
        this->mm_assign_module(i,DARWIN_MM_JOINTS);
      else if(module=="head")
        this->mm_assign_module(i,DARWIN_MM_HEAD);
      else if(module=="gripper")
        this->mm_assign_module(i,DARWIN_MM_GRIPPER);
      else
        this->mm_assign_module(i,DARWIN_MM_NONE);
      break;
    }
  }
}

mm_mode_t CDarwinRobot::mm_get_module(unsigned char servo_id)
{
  unsigned char value;

  if(servo_id>MAX_NUM_SERVOS)
  {
    // handle exceptions
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  // get the current value
  this->robot_device->read_byte_register(DARWIN_MM_MODULE_EN0+servo_id/2,&value);
  if(servo_id%2)// odd servo
    return (mm_mode_t)(value&0x07);
  else
    return (mm_mode_t)((value&0x70)>>4);
}

void CDarwinRobot::mm_load_config(std::string &filename)
{
  darwin_config_t::servo_iterator iterator;
  unsigned int i=0;
  struct stat buffer;

  if(stat(filename.c_str(),&buffer)==0)
  {
    try{
      std::auto_ptr<darwin_config_t> cfg(darwin_config(filename.c_str(), xml_schema::flags::dont_validate));
      for(iterator=cfg->servo().begin(),i=0;iterator!=cfg->servo().end();iterator++,i++)
        this->mm_assign_module(iterator->name(),iterator->motion_module());
    }catch(const xml_schema::exception& e){
      std::ostringstream os;
      os << e;
      throw CDarwinRobotException(_HERE_,os.str());
    }
  }
  else
    throw CDarwinRobotException(_HERE_,"Configuration file does not exist");
}

std::vector<double> CDarwinRobot::mm_get_servo_angles(void)
{
  int i=0;
  short int values[MAX_NUM_SERVOS]; 
  std::vector<double> angles(MAX_NUM_SERVOS);

  this->robot_device->read_registers(DARWIN_MM_SERVO0_CUR_POS_L,(unsigned char *)values,MAX_NUM_SERVOS*2);

  for(i=0;i<MAX_NUM_SERVOS;i++)
    angles[i]=((double)values[i])/128.0;

  return angles;
}

double CDarwinRobot::mm_get_servo_angle(unsigned char servo_id)
{
  unsigned short int value;
  double angle;
  
  if(servo_id>MAX_NUM_SERVOS)
  {
    / handle exceptions
    throw CDarwinRobotException(_HERE_,"Invalid servo identifier");
  }
  this->robot_device->read_word_register(DARWIN_MM_SERVO0_CUR_POS_L+servo_id/2,&value);
  angle=((double)value)/128.0;

  return angle;
}

void CDarwinRobot::mm_set_balance_gains(double knee,double ankle_pitch,double hip_roll,double ankle_roll)
{
  this->robot_device->write_byte_register(DARWIN_MM_KNEE_GAIN,(unsigned char)(knee*64.0));
  this->robot_device->write_byte_register(DARWIN_MM_ANKLE_PITCH_GAIN,(unsigned char)(ankle_pitch*64.0));
  this->robot_device->write_byte_register(DARWIN_MM_HIP_ROLL_GAIN,(unsigned char)(hip_roll*64.0));
  this->robot_device->write_byte_register(DARWIN_MM_ANKLE_ROLL_GAIN,(unsigned char)(ankle_roll*64.0));
}

void CDarwinRobot::mm_get_balance_gains(double *knee,double *ankle_pitch,double *hip_roll,double *ankle_roll)
{
  unsigned char value;

  this->robot_device->read_byte_register(DARWIN_MM_KNEE_GAIN,&value);
  *knee=value/64.0;
  this->robot_device->read_byte_register(DARWIN_MM_ANKLE_PITCH_GAIN,&value);
  *ankle_pitch=value/64.0;
  this->robot_device->read_byte_register(DARWIN_MM_HIP_ROLL_GAIN,&value);
  *hip_roll=value/64.0;
  this->robot_device->read_byte_register(DARWIN_MM_ANKLE_ROLL_GAIN,&value);
  *ankle_roll=value/64.0;
}

// motion action interface
void CDarwinRobot::action_load_page(unsigned char page_id)
{
  this->robot_device->write_byte_register(DARWIN_ACTION_PAGE,page_id);
}

unsigned char CDarwinRobot::action_get_current_page(void)
{
  unsigned char page_id;

  this->robot_device->read_byte_register(DARWIN_ACTION_PAGE,&page_id);

  return page_id;
}

void CDarwinRobot::action_start(void)
{
  this->robot_device->write_byte_register(DARWIN_ACTION_CONTROL,0x01);
}

void CDarwinRobot::action_stop(void)
{
  this->robot_device->write_byte_register(DARWIN_ACTION_CONTROL,0x02);
}

bool CDarwinRobot::action_is_page_running(void)
{
  unsigned char status;

  this->robot_device->read_byte_register(DARWIN_ACTION_STATUS,&status);
  if(status&0x01)
    return true;
  else
    return false;
}

// walking interface
void CDarwinRobot::walking_set_speeds(double fw_step,double side_step,double turn)
{
  this->robot_device->write_byte_register(DARWIN_STEP_FW_BW,(unsigned char)fw_step);
  this->robot_device->write_byte_register(DARWIN_STEP_LEFT_RIGHT,(unsigned char)side_step);
  this->robot_device->write_byte_register(DARWIN_STEP_DIRECTION,(unsigned char)(turn*8.0));
}

void CDarwinRobot::walking_start(void)
{
  this->robot_device->write_byte_register(DARWIN_WALK_CONTROL,0x01);
}

void CDarwinRobot::walking_stop(void)
{
  this->robot_device->write_byte_register(DARWIN_WALK_CONTROL,0x02);
}

bool CDarwinRobot::is_walking(void)
{
  unsigned char status;  

  this->robot_device->read_byte_register(DARWIN_WALK_STATUS,&status);
  if(status==0x01)
    return true;
  else
    return false;
}

void CDarwinRobot::walking_set_param(int param_id,double value)
{
  switch(param_id)
  {
    case DARWIN_X_OFFSET: 
    case DARWIN_Y_OFFSET: 
    case DARWIN_Z_OFFSET:
    case DARWIN_STEP_FW_BW: 
    case DARWIN_FOOT_HEIGHT: 
    case DARWIN_SWING_RIGHT_LEFT: 
    case DARWIN_SWING_TOP_DOWN: 
    case DARWIN_P_GAIN: 
    case DARWIN_I_GAIN: 
    case DARWIN_D_GAIN: 
    case DARWIN_STEP_LEFT_RIGHT: this->robot_device->write_byte_register(param_id,(char)value);
                                 break;
    case DARWIN_ROLL: 
    case DARWIN_PITCH: 
    case DARWIN_YAW: 
    case DARWIN_PELVIS_OFFSET: 
    case DARWIN_STEP_DIRECTION: this->robot_device->write_byte_register(param_id,(char)(value*8.0));
                                break;
    case DARWIN_HIP_PITCH_OFF_L: this->robot_device->write_word_register(param_id,(short int)(value*1024.0));
                                 break;
    case DARWIN_PERIOD_TIME_L: this->robot_device->write_word_register(param_id,(short int)(value));
                               break;
    case DARWIN_DSP_RATIO: 
    case DARWIN_STEP_FW_BW_RATIO: this->robot_device->write_byte_register(param_id,(char)(value*128.0));
                                  break;
    case DARWIN_ARM_SWING_GAIN: this->robot_device->write_byte_register(param_id,(char)(value*32.0));
                                break;
    case DARWIN_MAX_ACC: 
    case DARWIN_MAX_ROT_ACC: this->robot_device->write_byte_register(param_id,(char)(value*1024));
                             break;
  }
}

double CDarwinRobot::walking_get_param(int param_id)
{
  unsigned char byte_value;
  unsigned short int word_value;
  double value;

  switch(param_id)
  {
    case DARWIN_X_OFFSET: 
    case DARWIN_Y_OFFSET: 
    case DARWIN_Z_OFFSET:
    case DARWIN_STEP_FW_BW: 
    case DARWIN_FOOT_HEIGHT: 
    case DARWIN_SWING_RIGHT_LEFT: 
    case DARWIN_SWING_TOP_DOWN: 
    case DARWIN_P_GAIN: 
    case DARWIN_I_GAIN: 
    case DARWIN_D_GAIN: 
    case DARWIN_STEP_LEFT_RIGHT: this->robot_device->read_byte_register(param_id,&byte_value);
                                 value=(double)((char)byte_value);
                                 break;
    case DARWIN_ROLL: 
    case DARWIN_PITCH: 
    case DARWIN_YAW: 
    case DARWIN_PELVIS_OFFSET: 
    case DARWIN_STEP_DIRECTION: this->robot_device->read_byte_register(param_id,&byte_value);
                                value=((double)((char)byte_value))/8.0;
                                break;
    case DARWIN_HIP_PITCH_OFF_L: this->robot_device->read_word_register(param_id,&word_value);
                                 value=((double)((short int)word_value))/1024.0;
                                 break;
    case DARWIN_PERIOD_TIME_L: this->robot_device->read_word_register(param_id,&word_value);
                               value=((double)((short int)word_value));
                               break;
    case DARWIN_DSP_RATIO: 
    case DARWIN_STEP_FW_BW_RATIO: this->robot_device->read_byte_register(param_id,&byte_value);
                                  value=((double)((char)byte_value))/128.0;
                                  break;
    case DARWIN_ARM_SWING_GAIN: this->robot_device->read_byte_register(param_id,&byte_value);
                                value=((double)((char)byte_value))/32.0;
                                break;
    case DARWIN_MAX_ACC:
    case DARWIN_MAX_ROT_ACC: this->robot_device->read_byte_register(param_id,&byte_value);
                             value=((double)((char)byte_value))/1024.0;
                             break;
  }
  
  return value;
}

// joints interface
void CDarwinRobot::joints_load(std::vector<int> &servos,std::vector<double> &angles,std::vector<double> &speeds,std::vector<double> &accels)
{
  unsigned int i=0;
  
  for(i=0;i<servos.size();i++)
  {
    this->robot_device->write_word_register(DARWIN_MM_SERVO0_CUR_POS_L+servos[i]*2,((unsigned short int)(angles[i]*128.0)));  
    this->robot_device->write_byte_register(DARWIN_SERVO_0_SPEED+servos[i],(unsigned char)speeds[i]);  
    this->robot_device->write_byte_register(DARWIN_SERVO_0_ACCEL+servos[i],(unsigned char)accels[i]);  
  }
}

void CDarwinRobot::joints_start(void)
{
  this->robot_device->write_byte_register(DARWIN_JOINTS_CONTROL,0x01);
}

void CDarwinRobot::joints_stop(void)
{
  this->robot_device->write_byte_register(DARWIN_JOINTS_CONTROL,0x02);
}

bool CDarwinRobot::are_joints_moving(void)
{
  unsigned char status;  

  this->robot_device->read_byte_register(DARWIN_JOINTS_STATUS,&status);
  if(status==0x01)
    return true;
  else
    return false;
}

// head tracking interface
void CDarwinRobot::head_tracking_set_target(double pan,double tilt)
{
  this->robot_device->write_word_register(DARWIN_HEAD_PAN_L,(unsigned short int)(pan*128.0));
  this->robot_device->write_word_register(DARWIN_HEAD_TILT_L,(unsigned short int)(tilt*128.0));
}

void CDarwinRobot::head_tracking_set_pan_pid(double p,double i,double d)
{
  this->robot_device->write_byte_register(DARWIN_HEAD_PAN_P,(unsigned char)(p*1024.0));
  this->robot_device->write_byte_register(DARWIN_HEAD_PAN_I,(unsigned char)(i*1024.0));
  this->robot_device->write_byte_register(DARWIN_HEAD_PAN_D,(unsigned char)(d*1024.0));
}

void CDarwinRobot::head_tracking_set_tilt_pid(double p,double i,double d)
{
  this->robot_device->write_byte_register(DARWIN_HEAD_TILT_P,(unsigned char)(p*1024.0));
  this->robot_device->write_byte_register(DARWIN_HEAD_TILT_I,(unsigned char)(i*1024.0));
  this->robot_device->write_byte_register(DARWIN_HEAD_TILT_D,(unsigned char)(d*1024.0));
}

void CDarwinRobot::head_tracking_start(void)
{
  this->robot_device->write_byte_register(DARWIN_HEAD_CONTROL,0x01);
}

void CDarwinRobot::head_tracking_stop(void)
{
  this->robot_device->write_byte_register(DARWIN_HEAD_CONTROL,0x00);
}

bool CDarwinRobot::is_head_tracking(void)
{
  unsigned char data;

  this->robot_device->read_byte_register(DARWIN_HEAD_STATUS,&data);
  if(data==0x00)
    return false; 
  else
    return true;
}

/ gripper interface 
void CDarwinRobot::gripper_open(grippers_t gripper)
{
  if(gripper==LEFT_GRIPPER)
    this->robot_device->write_byte_register(DARWIN_GRIPPER_CONTROL,0x01);
  else
    this->robot_device->write_byte_register(DARWIN_GRIPPER_CONTROL,0x04);
}

void CDarwinRobot::gripper_close(grippers_t gripper)
{
  if(gripper==LEFT_GRIPPER)
    this->robot_device->write_byte_register(DARWIN_GRIPPER_CONTROL,0x02);
  else
    this->robot_device->write_byte_register(DARWIN_GRIPPER_CONTROL,0x08);
}*/

CDarwinRobot::~CDarwinRobot()
{
  if(this->robot_device!=NULL)
  {
    this->dyn_server->free_device(this->robot_id);
  }
}

